/*
	Double Frequency Modulation Music Synthesizer v1.5
	Copyright (C) 2000 Alexey Spiridonov

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream.h>
#include <math.h>
#include <string.h>
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>

#define NOTE_VALUE 512.0

#define ECHO ((long)(ECHO_DELAY * samplesPerSec))

#ifndef GO_AHEAD
#error Make sure 'value' is declared correctly before compiling, #define GO_AHEAD when done
#endif

typedef unsigned short int value;

const enum {Dampening,O1,I1,O2,DO1,FO,DO2,MAIN_VOL,EFFECT_VOL,MAX_OPTS};
const int noteBufferSize = 64;

struct Note {
  double balance;
  double freq;
  double duration;
  double wait;
  double volume;
	double params[MAX_OPTS];
 };

struct NoteBufferElement {
  Note *note;
  int inUse;
  double startedAt;
  long nextNote;
 };

double ECHO_VOL=0.1; //90% intensity drop
double VOL_MUL=0.909; //0.909 because users are likely to assume that the MAIN_VOL & EFFECT_VOL are out of 1.00
double ECHO_DELAY=0.3; //0.3 second delay
long BUFFER_SIZE = 131072; //Disk buffer size 

//A DFM-note producing function
value playNote(double Hertz, double Duration, double Loudness, double Position, double sampleRate, double params[MAX_OPTS])
{
	double x = (double(Position)*params[Dampening])/Duration, y = Hertz * Position * 3.14159265358979 / sampleRate;
	return value(Loudness * (NOTE_VALUE * sin(params[O1] * sin(y * params[I1]) + params[O2] * sin(y)) + NOTE_VALUE) * (exp(-pow(x + params[DO1], params[FO]) - x + params[DO2])));
}

//A cubic frequency interpolator to create smooth transitions between notes. Ends up producing
//percussion-like effects when used on several voices simultaneously. Marginally useful.
double transition(NoteBufferElement cur, Note next, double where)
{
	return ((6*(cur.note->freq - next.freq)/(cur.note->duration*cur.note->duration*cur.note->duration)) * ( ((where*where*where)/3) - (cur.note->duration*(where*where)/2) )) + cur.note->freq;
}

char *bob; //Name of the template file

void playMusic(ifstream &in, ofstream &out, long samplesPerSec);

int main(int argc, char **argv)
{
	//Take care of legalities in the most blunt possible way
	cout << "Double Frequency Modulation Music Synthesizer v1.5\n\
Copyright (C) 2000 Alexey Spiridonov\n\
\n\
This program is free software; you can redistribute it and/or modify\n\
it under the terms of the GNU General Public License as published by\n\
the Free Software Foundation; either version 2 of the License, or\n\
(at your option) any later version.\n\
\n\
This program is distributed in the hope that it will be useful,\n\
but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
GNU General Public License for more details.\n\
\n\
You should have received a copy of the GNU General Public License\n\
along with this program; if not, write to the Free Software\n\
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n";

	long sRate;
	ifstream options("synthopt.dat");

	if (options)
	{
		cout << "Loading options from synthopt.dat:\n";
		options >> ECHO_VOL >> VOL_MUL >> ECHO_DELAY >> BUFFER_SIZE;
		options.close();
	}
	else
		cout << "Using default options (use synthopt.dat to change):\n ";

	cout << "\tEcho volume - " << ECHO_VOL << ", Volume multiplier - " << VOL_MUL
	     << ",\n\tEcho delay - " << ECHO_DELAY << ", Disk buffer size - " << BUFFER_SIZE << "\n\n";
	if ((double(BUFFER_SIZE)/double(2*sizeof(value)) - double(BUFFER_SIZE/(2*sizeof(value)))) > 0.0000001)
	{
		cout << "Disk buffer size must be a multiple of " << (2*sizeof(value)) << ".\n";
		return 10;
	}

 	if (argc <= 3)
	{
	  	cout << "Usage: " << argv[0] << " miffile wavfile sampleRate \n	miffile format\n"
    		 << "\t<number of entries>\n\t<frequency> <duration> <wait> <volume> <balance> <"
        	 << MAX_OPTS << " instrument params>\n		. . .\n";
		return 1;
	}

	switch (argv[3][0])
	{
		case '4':
			sRate = 44100;
		  bob = "template44100s.wav";
		break;
		case '2':
		 	sRate = 22050;
			bob = "template22050s.wav";
		break;
		case '1':
			sRate = 11025;
			bob = "template11025s.wav";
		break;
		case '8':
			sRate = 8000;
			bob = "template8000s.wav";
		break;
		default:
			cout << "Unrecognized sample rate (use 44100, 22050, 11025, or 8000)." << "\n";
			return 3;
	}

	ifstream in(argv[1]);
	ofstream out(argv[2],ios::binary | ios::out);
	if (in == NULL)
	{
		cout << "Unable to open " << argv[1] << "\n";
		return 2;
	}
	if (out == NULL)
	{
		cout << "Unable to open " << argv[2] << "\n";
		return 2;
	}

	cout << "Synthesizing music according to MIF file at " << sRate << "kHz, stereo...\n";
	playMusic(in, out, sRate);
	cout << "DONE" << "\n";

	//Clean up
	in.close();
	out.close();
	return 0;
}

void playMusic(ifstream &in, ofstream &out, long samplesPerSec)
{
  //A little intialization
  char *buffer = new char[BUFFER_SIZE];
  long bufferPos = 0;
	long nNotes;
	in >> nNotes;

	Note *notes = new Note[nNotes];

	NoteBufferElement noteBuffer[noteBufferSize];

	double i, lastOver, duration = 0;
	value temp, r=0, l=0, prevR, prevL;
	double *rhistory= new double[ECHO], *lhistory = new double[ECHO];
	long historyPos = 0;
	long curnote = 0;
	long j,k, displayCount = 0;


	//Complete silence stood in the recording hall before synthesis began...
	for (j = 0; j < ECHO; ++j)
	{
		rhistory[j] = 0.0;
		lhistory[j] = 0.0;
	}


	//Load the notes from the file
	for (j = 0; j < nNotes; ++j)
	{
		in >> notes[j].freq >> notes[j].duration >> notes[j].wait >> notes[j].volume >> notes[j].balance;
		for (k = 0; k < MAX_OPTS; ++k)
			in >> notes[j].params[k];
		notes[j].duration = ((double)samplesPerSec*(double)notes[j].duration)/1000.0;
		notes[j].wait = ((double)samplesPerSec*(double)notes[j].wait)/1000.0;
		duration += notes[j].duration + notes[j].wait;
  }


	//Initialize the note buffer
	for (j=0; j < noteBufferSize; ++j)
		noteBuffer[j].inUse = 0;

	cout << "\tSamples to generate: " << duration << " \n";


	//We need to load most of the RIFF header from a template file; I hacked the
	//WAV format by hand in a hurry and only know a few positions to modify to good
	//effect
	cout << "\tMaking RIFF header..." << flush;
	ifstream wave(bob,ios::binary | ios::in);
	if (!wave)
	{
		cout << "Unable to open " << bob << "\n";
		return;
	}

	for (i = 0; i < 4; ++i)
		out << char(wave.get());
	for (i = 0; i < 4; ++i)
		wave.get();

  //Bad variable names, but we won't need the other bob ;)
  //Don't you love C++ scoping rules?
	long bob = long(2*duration*sizeof(value)+50);
	out.write((char*)&bob,sizeof(long));

	for (i = 0; i < 38; ++i)
		out << char(wave.get());
	for (i = 0; i < 4; ++i)
		wave.get();

	bob = long(duration);
	out.write((char*)&bob,sizeof(long));

	for (i = 0; i < 4; ++i)
		out << char(wave.get());
	for (i = 0; i < 4; ++i)
		wave.get();

	bob = long(2*duration*sizeof(value));
	out.write((char*)&bob,sizeof(long));

	wave.close();
	cout << "DONE" << "\n" << flush;


	//Now comes the core
	cout << "\tGenerating samples" << flush;

	for (i = 0; i < duration; ++i)
	{
  	//Progress indicator
		if ((unsigned long)(i) % 10000 == 0)
    {
    	if (displayCount == 0)
      	cout << "\n\t\t[";
			cout << "." << flush;
      if (displayCount == 49)
      	cout << "]";
      displayCount = (displayCount + 1) % 50;
    }
    //Add as many notes to buffer as we need to (if we can)
		do
		{
      //Find the note that will finish playing last.
			lastOver = i;
			for (j = 0; j < noteBufferSize; ++j)
				if (noteBuffer[j].inUse && (noteBuffer[j].startedAt + noteBuffer[j].note->duration + noteBuffer[j].note->wait) > lastOver)
					lastOver = noteBuffer[j].startedAt + noteBuffer[j].note->duration + noteBuffer[j].note->wait;
      //If it just finished, add the next one in
			if (lastOver == i)
			{
				j = 0;
        //Find a free slot in the buffer, if available
				while (j < noteBufferSize && noteBuffer[j].inUse == 1) j++;
        //Place the next note in the free slot (skip if unavailable)
				if (j < noteBufferSize  && curnote < nNotes)
				{
					noteBuffer[j].note = &notes[curnote];

					noteBuffer[j].startedAt = i;
					noteBuffer[j].inUse = 1;
					if (curnote != nNotes - 1)
						noteBuffer[j].nextNote = curnote + 1;
					else
						noteBuffer[j].nextNote = curnote;
					curnote++;
				}
			}
		} while ((lastOver <= i) && (j < noteBufferSize) && (curnote < nNotes));
    //We keep the previous value to avoid sharp skips
		prevR = r;
		prevL = l;
		r = 0;
		l = 0;
    //Calculate the waveform for each of the notes in turn, and add to the totals
		for (j = 0; j < noteBufferSize; ++j)
		{
    	//Remove note from buffer if done playing
			if ((noteBuffer[j].inUse) && ((noteBuffer[j].startedAt + noteBuffer[j].note->duration + ((noteBuffer[j].note->wait > 0) ? noteBuffer[j].note->wait : 0)) < i) )
				noteBuffer[j].inUse = 0;
      //Otherwise (with a small additional (necessary, as far as I remember) check) calculate the note's waveform
			if (noteBuffer[j].inUse && noteBuffer[j].startedAt + noteBuffer[j].note->duration > i)
			{
				temp = 0;
				//Save the calculation time the effect or the note is turned off
				if (noteBuffer[j].note->params[MAIN_VOL] > 0.0001)
					temp += (value)(noteBuffer[j].note->params[MAIN_VOL] * playNote(noteBuffer[j].note->freq, noteBuffer[j].note->duration, noteBuffer[j].note->volume, i - noteBuffer[j].startedAt, samplesPerSec, noteBuffer[j].note->params));
				if (noteBuffer[j].note->params[EFFECT_VOL] > 0.0001)
					temp += (value)(noteBuffer[j].note->params[EFFECT_VOL] * playNote(transition(noteBuffer[j], notes[noteBuffer[j].nextNote],i), noteBuffer[j].note->duration, noteBuffer[j].note->volume, i - noteBuffer[j].startedAt, samplesPerSec, noteBuffer[j].note->params));
        //Add the properly stereo shifted waveform to the totals
				r += (value)(2 * noteBuffer[j].note->balance * temp);
				l += (value)(2 * (1 - noteBuffer[j].note->balance) * temp);
			}
		}
		//Remove the sharp skip between a high plateaue and a drop to no notes playing
    //This sounds as a pop otherwise
		if (fabs(r+l) <= 0.001)
		{
			r = prevR;
			l = prevL;
		}
    //Add and save the echo.
		r += (value)(ECHO_VOL * rhistory[historyPos]);
		l += (value)(ECHO_VOL * lhistory[historyPos]);
		rhistory[historyPos] = r;
		lhistory[historyPos] = l;
		historyPos = (historyPos + 1) % long(ECHO);
    //Write the right and left channels to a buffer (and optionally dump the buffer to a file)
		memcpy((void*)(buffer + bufferPos), (void*)(&r), sizeof(value));
		bufferPos = (bufferPos + sizeof(value)) % BUFFER_SIZE;
		memcpy((void*)(buffer + bufferPos), (void*)(&r), sizeof(value));
		bufferPos = (bufferPos + sizeof(value)) % BUFFER_SIZE;
		if (bufferPos == 0)
			out.write(buffer, BUFFER_SIZE);
	}
	//Flush the buffer
  out.write(buffer, bufferPos);
	//Close progress indicator bracket if we hadn't already
  if (displayCount != 0)
  	cout << "]" << flush;
  //Done!
	cout << "\n\tDONE\n" << flush;;

  //Clean up
	delete [] notes;
	delete [] rhistory;
	delete [] lhistory;
}
